# -*- coding: utf-8 -*-
"""Support the research during the annual Lehrexkursion at Universität Hamburg.
"""
from os.path import dirname, join

from .csv import *
from . import plots
from . import utils


__all__ = [s for s in dir() if not s.startswith('_')]
__version__ = open(join(dirname(__file__), 'VERSION')).read().strip()



# -*- coding: utf-8 -*-
"""Plot basic atmospheric properties conveniently.
"""
from lehrex.plots.common import *  # noqa
from lehrex.plots.plots import *  # noqa


__all__ = [s for s in dir() if not s.startswith('_')]



# -*- coding: utf-8 -*-
"""Load CSV files stored in RAW format.
"""
import re
from warnings import warn
from datetime import datetime

import pandas as pd


__all__ = [
    'read_raw',
]


def _parse_starttime(filename):
    """Parse start time from file header and return date string."""
    with open(filename, 'rt', encoding='UTF-8', errors='ignore') as f:
        header = f.readline()

    return re.compile('.*: (.*) UTC.*').match(header).group(1)


def _get_date_range(filename, freq, periods, datefmt='%d.%m.%Y %H:%M:%S,%f'):
    """Return array of datetime objects from file."""
    start = datetime.strptime(_parse_starttime(filename), datefmt)

    return pd.date_range(start, freq=freq, periods=periods)


def read_raw(filename, freq='50ms', decimal=',', delimiter=';', comment='%',
             header=0, **kwargs):
    """Read file in RAW format.

    Parameters:
        filename (str): Path to CSV file.
        freq (str or pd.DateOffset): Data frequency see `pd.read_csv`.
        decimal (str): Character to recognize as decimal point.
        delimiter (str): The string used to separate values.
        comment (str): The character used to indicate a comment line.
        header (int): Row number to use as the column names.
            See `pd.read_csv` for further information.
        **kwargs: Additional keyword arguments passed to `pd.read_csv`.

    Returns:
        dict: Dictionary containing the data arrays.
    """
    df = pd.read_csv(
        filename,
        comment=comment,
        decimal=decimal,
        delimiter=delimiter,
        header=header,
        **kwargs,
    )

    try:
        # Try to parse the start time from file header.
        df['DATETIME'] = _get_date_range(
            filename, freq=freq, periods=df.shape[0])
    except ValueError:
        # Do not set index, if the regex in `_get_date_range` does not match.
        warn('Could not parse dateformat in file header. No data index set.')
    else:
        # Use date range as index if it is parsed correctly.
        df = df.set_index('DATETIME')

    return df



# -*- coding: utf-8 -*-
"""Common variables and functions.
"""
__all__ = [
    'get_var_desc',
    'get_label',
]


_variable_description = {
    'AH002': ('Absolute Feuchte 2 m', 'g/m³'),
    'ALB': ('Albedo', '1 '),
    'D': ('Diffuse Himmelsstrahlung', 'W/m²'),
    'DD010': ('Windrichtung 10 m', '°'),
    'DR': ('Diffuse Himmelsstrahlung (unkorrigiert)', 'W/m²'),
    'DT002': ('Taupunkt 2 m', '°C'),
    'E': ('Langwellige Strahlung von unten', 'W/m²'),
    'ETS': ('Oberflächentemperatur (mit EPS aus Standortparametern)', '°C'),
    'FB010': ('Stärkste Böen 10 m', 'm/s'),
    'FF010': ('Windgeschwindigkeit 10 m', 'm/s'),
    'G': ('Globalstrahlung', 'W/m²'),
    'GD': ('Diffuse Himmelsstrahlung (aus Globalstrahlung)', 'W/m²'),
    'GI': ('Direkte Sonnenstrahlung (aus Globalstrahlung)', 'W/m²'),
    'GND': ('Sonnenschein nicht möglich (Nacht)', ''),
    'GP': ('Relative Globalstrahlung', '%'),
    'GSD': ('Sonnenscheindetektion', ''),
    'GSH': ('Sonnenscheindauer', 'h'),
    'GSPT': ('Relative Sonnenscheindauer bzgl. 1 Tag', '%'),
    'GSPX': ('Relative Sonnenscheindauer', '%'),
    'GSW': ('Sonnenscheinschwellwert', 'W/m²'),
    'GSZ': ('Summierte Sonnenscheindetektionen', '1'),
    'GTD': ('Schattendetektion', ''),
    'GTH': ('Schattendauer', 'h'),
    'GTZ': ('Summierte Schattendetektionen', '1'),
    'GXD': ('Sonnenschein möglich', ''),
    'GXH': ('Mögliche Sonnenscheindauer', 'h'),
    'GXT': ('Max. mögliche Tagessonnenscheindauer', 'h'),
    'GXZ': ('Summierter möglicher Sonnenschein', '1'),
    'HTD600': ('HMP-Mast Taupunkt, 6 m', '°C'),
    'HTT050': ('HMP-Mast Lufttemperatur, 0,5 m', '°C'),
    'HTT200': ('HMP-Mast Lufttemperatur, 2 m', '°C'),
    'HTT600': ('HMP-Mast Lufttemperatur, 6 m', '°C'),
    'I': ('Direkte Sonnenstrahlung', 'W/m²'),
    'IC': ('Direkte Sonnenstrahlung bei wolkenlosen Bedingungen', 'W/m²'),
    'L': ('Langwellige Strahlung von oben', 'W/m²'),
    'LTS': ('Himmelstemperatur (mit EPS = 1)', '°C'),
    'MG': ('Theoretische Globalstrahlung bei wolkenlosem Himmel', 'W/m²'),
    'MH002': ('Massenmischungsverhältnis 2 m', 'g/kg'),
    'P000': ('Luftdruck (Meereshöhe)', 'hPa'),
    'P007': ('Luftdruck (Stationshöhe)', 'hPa'),
    'PBA': ('Steigwinkel Pilotballon', '°'),
    'PBG': ('Füllgewicht Pilotballon', 'g'),
    'PBW': ('Steiggeschwindigkeit Pilotballon', 'm/min'),
    'Q': ('Strahlungsbilanz', 'W/m²'),
    'R': ('Kurzwellige Strahlung von unten', 'W/m²'),
    'RH002': ('Relative Feuchte 2 m', '%'),
    'RK': ('Niederschlagsmenge ab 0 Uhr', 'mm'),
    'RR': ('Niederschlagsintensität', 'mm'),
    'SH002': ('Spezifische Feuchte 2 m', 'g/kg'),
    'SOLH': ('Höhenwinkel der Sonne', '°'),
    'TT002': ('Lufttemperatur 2 m', '°C'),
    'VP002': ('Wasserdampfdruck 2 m', 'hPa'),
}


def get_var_desc():
    """Get a copy of the default variable description.

    Returns:
        dict: Dict values are tuples containing full variable name and unit.
            The keys are the abbrevations used in `MASTER.txt`.

    Examples:
        Add description for non-default variables:
        >>> desc = lx.plots.get_var_desc()
        >>> desc.update({'FOO': ('New variable', 'Unit')})
        >>> lx.plots.get_label('FOO', var_desc=desc)
        'New variable [Unit]'
    """
    return _variable_description.copy()


def get_label(key, label='{name} [{unit}]', var_desc=None):
    """Return label for variable key.

    Parameters:
        key (str): Variable key.
        label (str): Format string to create the label.
            The variables `name` and `unit` can be used.
        var_desc (dict): Dictionary with variable descriptions.
            The value behind the key has to be a tuple of strings:
                dict[key] = (name, unit)
            If `None` a default set of variables is used.
    Returns:
        str: Variable specific label.

    Examples:
        Get default axis label for variable key "TT002":
        >>> get_label('TT002')
        'Lufttemperature 2 m [K]'

        Pass different format string for label:
        >>> get_label('TT002', label='{name} in {unit}')
        'Lufttemperatur in 2 m in K'

        Set description for non-default variables:
        >>> desc = lx.plots.get_var_desc()
        >>> desc.update({'FOO': ('New variable', 'Unit')})
        >>> get_label('FOO', var_desc=desc)
        'New variable [Unit]'
    """
    if var_desc is None:
        var_desc = _variable_description

    name, unit = var_desc.get(key, (key, ''))

    kwargs = {
        'name': name,
        'unit': unit,
    }
    return label.format(**kwargs)



# -*- coding: utf-8 -*-
"""Plot basic atmospheric properties conveniently.
"""
from matplotlib import dates
import matplotlib.pyplot as plt

from .common import get_label


__all__ = [
    'set_date_axis',
    'timeseries',
    'timeseries2d',
]


def set_date_axis(ax=None, dateformat='%d.%m.'):
    """Set DateFormatter for given AxesSubplot.

    Parameters:
        ax (AxesSubplot): Matplotlib axes.
        dateformat (str): Format string for date labels on xaxis.
    """
    if ax is None:
        ax = plt.gca()

    formatter = dates.DateFormatter(dateformat)
    ax.xaxis.set_major_formatter(formatter)
    ax.grid('on', which='both', linestyle='-')


def timeseries(dataframe, key, xlabel='Datum', label=None, dateformat='%d.%m.',
               ax=None, **kwargs):
    """Create a basic timeseries plot.

    Parameters:
        dataframe (pd.Dataframe): Dataframe.
        key (str): Variable key.
        xlabel (str): x label.
        label (str): Line label.
        dateformat (str): Format string for date labels on xaxis.
        ax (AxesSubplot): Matplotlib axes.
        **kwargs: Additional keyword arguments passed to `plt.plot`.

    Returns:
        Line2D: A line.
    """
    if ax is None:
        ax = plt.gca()

    if label is None:
        label = get_label(key, label='{name}')

    line, = ax.plot(dataframe.index, dataframe[key], label=label, **kwargs)
    set_date_axis(ax, dateformat)

    ax.set_xlabel(xlabel)
    ax.legend()

    return line


def timeseries2d(x, y, c, xlabel='Datum', ylabel='Höhe [m]',
                 zlabel='', dateformat='%d.%m.', ax=None, **kwargs):
    """Create a basic 2D timeseries plot.

    Parameters:
        x, y, c (ndarray): Data arrays.
        xlabel (str): x label.
        ylabel (str): y label.
        zlabel (str): Colorbar label.
        dateformat (str): Format string for date labels on xaxis.
        ax (AxesSubplot): Matplotlib axes.
        **kwargs: Additional keyword arguments passed to `plt.pcolormesh`.

    Returns:
        matplotlib.collections.QuadMesh
    """
    if ax is None:
        ax = plt.gca()

    ret = ax.pcolormesh(x, y, c, **kwargs)

    cb = ax.get_figure().colorbar(ret)
    cb.set_label(zlabel)

    set_date_axis(ax, dateformat)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    return ret



# -*- coding: utf-8 -*-
"""General utility functions.
"""
import re

import numpy as np


__all__ = [
    'stack_series',
]


def stack_series(dataframe, regex=None):
    """Stack all Series in the DataFrame that match a given regular expression.

    This function converts data series into a multidimensional array (e.g.
    time series of height resolved data).

    Parameters:
        dataframe (pd.DataFrame): Dataframe.
        regex (str): Python regular expression [0] matching
            the variable names of the profile.

    [0] https://docs.python.org/3.1/library/re.html

    Returns:
        ndarray, ndarray: data array, height levels.

    Examples:
        >>> import lehrex as lex
        >>> df = lex.read('CLB.txt')
        >>> p, z = lex.utils.stack_series(df, regex='CLB_B\d{5}')

    """
    # Find all variable keys matching the given regular expression.
    pattern = re.compile(regex)
    var_names = list(filter(pattern.match, dataframe.keys()))

    profile = np.vstack([dataframe[v] for v in sorted(var_names)])

    # Retrieve height information from variable names.
    z = [float(re.sub('[^0-9]', '', v)) for v in var_names]

    return np.ma.masked_invalid(profile), np.array(z)



# -*- coding: utf-8 -*-
"""Load CSV files stored in Wettermast format.
"""
from collections import Iterable
from glob import iglob
from datetime import datetime

import numpy as np
import pandas as pd
from matplotlib.dates import strpdate2num


__all__ = [
    'read',
    'read_mfdataset',
    'write',
]


def _get_mpl_date(dates, fmt='%d.%m.%Y %H:%M'):
    """Convert date strings into matplotlib time format.

    Parameters:
        dates (ndarray): Array containing date strings.
        fmt (str): Date string format [0].

    [0] http://pubs.opengroup.org/onlinepubs/009695399/functions/strptime.html

    Returns:
        np.array: Matplotlib time values.
    """
    return np.array([strpdate2num(fmt)(d) for d in dates])


def _get_datetime(dataframe):
    """Convert 'DATE' and 'TIME' columns into list of datetime objects."""
    return [datetime.strptime(' '.join([d, t]), '%d.%m.%Y %H:%M')
            for d, t in zip(dataframe['DATE'], dataframe['TIME'])]


def _get_names(filename):
    """Return list of variable names."""
    with open(filename, "rb") as f:
        for line in f:
            if line.decode().startswith('$Names='):
                return line.decode().split('=')[1].strip().split(';')
    raise Exception('No field names were found.')


def _get_skip_header(filename):
    """Return number of lines to skip at the beginning of the file."""
    with open(filename, 'rb') as f:
        for i, line in enumerate(f):
            if not line.decode().startswith(('$', '#')):
                # Return number of first non-comment line.
                return i

    raise Exception('No valid line found.')


def read(filename, delimiter=';', **kwargs):
    """Read CSV file into a `pd.DataFrame`.

    Parameters:
        filename (str): Path to CSV file.
        delimiter (str): The string used to separate values.
        **kwargs: Additional keyword arguments passed to `pd.read_csv`.

    Returns:
        pd.DataFrame: DataFrame representing the timeseries.
    """
    # Parse CSV file and create `pandas.DataFrame`.
    df = pd.read_csv(
        filename,
        delimiter=delimiter,
        skiprows=_get_skip_header(filename),
        names=_get_names(filename),
        **kwargs,
    )

    # Convert 'DATE' and 'TIME' columns into datetime objects and use them
    # as DataFrame index.
    df['DATETIME'] = _get_datetime(df)
    df = df.set_index('DATETIME')

    return df


def read_mfdataset(file):
    """Read multiple CSV files and merge their data into one `pd.DataFrame`.

    Parameters:
       file (str or iterable): Either a string glob in the form
            “path/to/my/files/*.csv” or an iterable of files to read.

    Returns:
        pd.DataFrame: DataFrame storing the concatenated timeseries.
    """
    if isinstance(file, str):
        fiterator = iglob(file)
    elif isinstance(file, Iterable):
        fiterator = file
    else:
        raise TypeError('`file` has to be a glob string or iterable of paths.')

    return pd.concat(map(read, fiterator)).sort_index()


def write(filename, data, variables=None):
    """Write data from dictionary to CSV file.

    Parameters:
        filename (str): Path to CSV file.
        data (pd.Dataframe or dict-like): DataFrame or dict-like.
        variables (list[str]): Variables to store.
    """
    if not ('DATE' in data and 'TIME' in data):
        raise Exception('Need DATE and TIME column to write proper files.')

    if variables is None:
        variables = list(sorted(data.keys()))

    # Write date and time to first columns.
    for k in ['TIME', 'DATE']:
        if k in variables:
            variables.remove(k)
        variables.insert(0, k)

    # Convert dictionart entries to data matrix.
    data_matrix = np.vstack(data[v] for v in variables).T

    # Calculate some time statistics in Wettermast style.
    # It looks messy and it is not clear, if those values are needed.
    # But the results arer equivalent to the original files.
    base_date = '27.03.1995 00:00:00'
    dates = [' '.join(d) for d in zip(data['DATE'], data['TIME'])]
    first_date = dates[0] + ':00'
    mpl_times = _get_mpl_date([base_date, first_date], fmt='%d.%m.%Y %H:%M:%S')
    secs_since = np.diff(mpl_times) * 24 * 3600  # time span in seconds
    time_lag = np.round(np.mean(np.diff(_get_mpl_date(dates) * 24 * 3600)))

    # Combine time properties to file header.
    header = '#={}\n'.format(data_matrix.shape[0])
    header += '$FirstDateTime={}\n'.format(first_date)
    header += '$JSDBaseDateTime={}\n'.format(base_date)
    header += '$FirstJSD={}\n'.format(int(secs_since))
    header += '$TimeLagSec={}\n'.format(int(time_lag))
    header += '$DefaultValue=99999\n'
    header += '$Names=' + ";".join(variables)

    np.savetxt(
        filename,
        data_matrix,
        comments='',
        delimiter=';',
        header=header,
        fmt='%s',
    )



# -*- coding: utf-8 -*-
import sys
from distutils.core import setup
from os.path import dirname, join
from setuptools import find_packages


if not sys.version_info >= (3, 5, 1):
    sys.exit('Only support Python version >=3.5.1.\n'
             'Found version is {}'.format(sys.version))

__version__ = open(join(dirname(__file__), 'lehrex', 'VERSION')).read().strip()

setup(
    name='lehrex',
    author='Lukas Kluft',
    author_email='lukas.kluft@gmail.com',
    url='https://github.com/lkluft/lehrex',
    download_url='https://github.com/lkluft/lehrex/tarball/v' + __version__,
    version=__version__,
    packages=find_packages(),
    license='MIT',
    description='Support the research during the Lehrexkursion.',
    classifiers=[
        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    include_package_data=True,
    install_requires=[
        'matplotlib>=1.5.1',
        'numpy>=1.10.4',
        'nose',
        'pandas',
    ],
)



# -*- coding: utf-8 -*-
"""Basic I/O functionality for files in Wettermast format.
"""
from lehrex.csv.common import *  # noqa
from lehrex.csv.raw import *  # noqa


__all__ = [s for s in dir() if not s.startswith('_')]
